import time
from collections import OrderedDict

from pathlib import Path

import torch
from torch.autograd import Variable
from torch.utils.data import DataLoader
from ukro_g2p.predict import _collate_fn
from torch.optim.lr_scheduler import ReduceLROnPlateau

# from text import text_to_sequence
from ukro_g2p.predict import infolog
from tensorboardX import SummaryWriter
from tqdm import tqdm

log = infolog.log


class Trainer(object):
    def __init__(self, model, datasets, optimizer, loss, epochs, batch_size, experiment_name, logging_freq,
                 restore_epoch=-1):

        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.epochs = epochs
        self.experiment_name = experiment_name
        self.logging_freq = logging_freq
        self.model = model.to(self.device)
        self.restore_epoch = restore_epoch

        self.optimizer = optimizer
        self.loss = loss
        # TODO remove scheduler from trainer object
        self.scheduler = ReduceLROnPlateau(self.optimizer, mode='min', factor=0.1, patience=5, verbose=True)

        self.global_batch_index = 0
        self.global_batch_index_dev = 0
        self.datasets = datasets
        self.dataloaders = OrderedDict()
        for data_name in datasets:
            self.dataloaders[data_name] = DataLoader(datasets[data_name], batch_size=batch_size,
                                                     shuffle=True if data_name == 'train' else False,
                                                     collate_fn=_collate_fn,
                                                     num_workers=8)


        self.experiment_dir = Path('./exp') / self.experiment_name
        self.log_dir = self.experiment_dir / 'logs'
        self.log_path = self.log_dir / 'log.txt'
        self.checkpoint_dir = self.experiment_dir / 'checkpoints'

        self.experiment_dir.mkdir(parents=True, exist_ok=True)
        self.log_dir.mkdir(parents=True, exist_ok=True)
        self.checkpoint_dir.mkdir(parents=True, exist_ok=True)

        # initialize logger
        infolog.init(self.log_path, self.experiment_name)
        # initialize TensorBoard writer
        self.writer = SummaryWriter(log_dir=self.log_dir)

        if self.restore_epoch != -1:
            self.load_from_checkpoint(epoch=self.restore_epoch)

    def train_and_validate(self):
        for e in range(self.restore_epoch + 1, self.epochs):
            for ds, dataloader in self.dataloaders.items():
                if ds != 'train':
                    self.model.eval()  # turn off batchnorm /dropout
                self.run_epoch(dataloader, dataset_name=ds, epoch=e)
                self.model.train()  # turn on batchnorm /dropout
            self.save_to_checkpoint(epoch=e)

    def run_epoch(self, dataloader, dataset_name, epoch):
        begin_time = time.time()
        train = dataset_name == 'train'
        num_batches = len(dataloader)
        avg_loss = 0

        progress_bar = tqdm(total=num_batches)
        for batch_idx, batch in enumerate(dataloader):
            word_inputs, input_length, pron_targets, p_length = batch

            word_inputs = Variable(word_inputs).to(self.device)
            input_length = Variable(input_length).to(self.device)
            pron_targets = Variable(pron_targets).to(self.device)
            p_length = Variable(p_length).to(self.device)

            p_preds, _ = self.model(word_inputs, input_length, pron_targets[:, :-1], p_length - 1)
            targets = pron_targets[:, 1:].contiguous()

            loss = self.loss(p_preds.view(p_preds.size(0) * p_preds.size(1), -1),
                             targets.view(targets.size(0) * targets.size(1)))

            loss_floats = loss.data.cpu().item()
            avg_loss += loss_floats

            progress_bar.set_description(f'{dataset_name}: {epoch}/{self.epochs}, current loss: {round(loss_floats,4)}')
            progress_bar.refresh()
            progress_bar.update()

            if train:
                self.global_batch_index += 1
                self.writer.add_scalar(dataset_name, loss_floats, self.global_batch_index)
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()
                # if (batch_idx + 1) % self.logging_freq == 0:
                #     log_str = self.construct_logging_str(loss_floats, epoch, num_batches, batch_idx + 1)
                #     log(log_str)
            else:
                # TODO it is ugly part of code, shuold add global step explicitly
                self.global_batch_index_dev += 1
                self.writer.add_scalar(dataset_name, loss_floats,  self.global_batch_index_dev)

        progress_bar.close()

        # add generated text to tensor board
        if not train:
            graphemes, graphemes_length, phonemes, _ = next(iter(dataloader))

            graphemes = graphemes.to(self.device)
            graphemes_length = graphemes_length.to(self.device)
            phonemes = phonemes.to(self.device)

            # phonemes_predictions = self.model(graphemes, graphemes_length).tolist()
            phonemes_targets = phonemes[:, 1:].contiguous().tolist()

            text = ''
            for idx in range(0, 5):
                phonemes_predictions = self.model(graphemes[idx].unsqueeze(0), graphemes_length[idx].unsqueeze(0)).tolist()[0]

                graphemes_str = ''.join(
                    [g for g in self.datasets[dataset_name].idx2graphemes(graphemes[idx].tolist()) if g not in ['<s>', 'pad']])
                predictions_str = ' '.join(
                    [p for p in self.datasets[dataset_name].idx2phonemes(phonemes_predictions) if p not in ['<os>', '</os>']])
                targets_str = ' '.join(
                    [p for p in self.datasets[dataset_name].idx2phonemes(phonemes_targets[idx]) if p not in ['<os>', '</os>', 'pad']])

                text = text + graphemes_str + '  \n' + targets_str + '  \n' + predictions_str + '  \n'

            self.writer.add_text('Text', text, epoch)

        avg_loss /= num_batches
        end_time = time.time()
        if train:
            self.scheduler.step(avg_loss)
        log_str = 'Epoch: {}, {} loss: {:.5f} time: {:.2f} sec'
        log_str = log_str.format(epoch, dataset_name, avg_loss, end_time - begin_time)
        log(log_str)

    def load_from_checkpoint(self, epoch):
        checkpoint_file = f'{self.experiment_name}-epoch-{epoch}.th'
        checkpoint_path = self.checkpoint_dir / checkpoint_file
        self.model.load_state_dict(torch.load(checkpoint_path, map_location=lambda storage, loc: storage))
        print('model loaded from checkpoint: {}'.format(checkpoint_file))

    def save_to_checkpoint(self, epoch):
        checkpoint_file = f'{self.experiment_name}-epoch-{epoch}.th'
        checkpoint_path = self.checkpoint_dir / checkpoint_file
        if self.checkpoint_dir.exists():
            torch.save(self.model.state_dict(), checkpoint_path)
            print(f'model saved to checkpoint: {checkpoint_path}')
        else:
            raise FileNotFoundError

    @staticmethod
    def construct_logging_str(loss, epoch, total_batches, idx):
        tmpstr = 'Epoch:{:2} Batch:[{:3}/{:3}] Loss: {:.4f}'
        tmpstr = tmpstr.format(epoch, idx, total_batches, loss)
        return tmpstr



import argparse
import torch
from torch.autograd import Variable

from ukro_g2p.models.g2p_model import G2PModel, pretrained_models
from ukro_g2p.tokenization import G2PTokenizer


class G2P(object):
    def __init__(self, model_name):
        super().__init__()

        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        self.model = G2PModel.from_pretrained(model_name).to(self.device)
        self.tokenizer = G2PTokenizer.from_pretrained(model_name)
        self.model.eval()

    def __call__(self, word, human_readable=False):

        # map word's chars into graphemes idx
        graphemes = self.tokenizer.tokenize_graphemes(word)
        graphemes = self.tokenizer.convert_graphemes_to_ids(graphemes)
        g_length = [len(graphemes)]

        graphemes = Variable(torch.LongTensor(graphemes).unsqueeze(0)).to(self.device)
        g_length = Variable(torch.LongTensor(g_length)).to(self.device)

        phonemes = self.model(graphemes, g_length).tolist()[0]

        return self.tokenizer.convert_ids_to_phonemes(phonemes) if not human_readable else self.tokenizer.convert_ids_to_human_phonemes(phonemes)


if __name__ == '__main__':
    # argument parser
    parser = argparse.ArgumentParser()

    # required
    parser.add_argument('word', type=str, help='Word to generate pronunciation for')
    parser.add_argument('-m', '--model_name', required=False, type=str, default='ukro-base-uncased',
                        choices=pretrained_models.keys(),
                        help='Trained model name')
    parser.add_argument('-hr', '--human_readable', required=False, action='store_true',
                        help='Output human readable set of phonemes')

    # parse
    script_args = parser.parse_args()

    g2p = G2P('ukro-base-uncased')
    pron = g2p(script_args.word, human_readable=True if script_args.human_readable else False)

    print(f"{' '.join(pron)}")



import torch.nn as nn
import torch
from torch.autograd import Variable
import numpy as np
from ukro_g2p.models.modules import Encoder, Decoder
from ukro_g2p.models.modules import Beam
from pathlib import Path
import configparser
from collections import namedtuple
from torch.utils import model_zoo

model_obj = namedtuple("model_obj", ["url", "config_path"])

pretrained_models = {
    "ukro-base-uncased": model_obj(
        url="https://github.com/kosti4ka/ukro_g2p/releases/download/ukro_base_uncased_v.0.1/ukro_base_uncased-epoch-99-d545c0d.th",
        config_path=Path(__file__).parent / "../configs/ukro_base_uncased.config",
    )
}


class G2PConfig(dict):

    def __init__(self, model_config_file):
        super(G2PConfig, self).__init__()

        self.use_cuda = False
        self.model_path = None

        # reading config file
        config_file = configparser.ConfigParser()
        config_file.read(model_config_file, encoding='utf8')

        self.padding = config_file['VocabConfig']['padding']
        # TODO simplify this part - use same bos, eos symbol for both phonemes and graphemes
        self.decoder_bos = config_file['VocabConfig']['phoneme_bos']
        self.decoder_eos = config_file['VocabConfig']['phoneme_eos']
        # reading graphemes
        self.graphemes = config_file['VocabConfig']['graphemes'].split()
        self.encoder_vocab_size = len(self.graphemes)
        self.encoder_padding_idx = self.graphemes.index(self.padding)
        # reading phonemes
        self.phonemes = config_file['VocabConfig']['phonemes'].split()
        self.decoder_vocab_size = len(self.phonemes)
        self.decoder_padding_idx = self.phonemes.index(self.padding)
        self.decoder_bos_idx = self.phonemes.index(self.decoder_bos)
        self.decoder_eos_idx = self.phonemes.index(self.decoder_eos)
        # reading human phonemes
        self.human_phonemes = config_file['VocabConfig']['human_phonemes'].split()

        # encoder config
        self.encoder_d_embed = int(config_file['EncoderConfig']['encoder_d_embed'])
        self.encoder_d_hidden = int(config_file['EncoderConfig']['encoder_d_hidden'])
        self.encoder_n_layers = int(config_file['EncoderConfig']['encoder_n_layers'])
        self.encoder_bidirectional = True if config_file['EncoderConfig']['encoder_bidirectional'].lower() == 'true' else False

        # decoder config
        self.decoder_d_embed = int(config_file['DecoderConfig']['decoder_d_embed'])
        self.decoder_d_hidden = int(config_file['DecoderConfig']['decoder_d_hidden'])
        self.decoder_n_layers = int(config_file['DecoderConfig']['decoder_n_layers'])
        self.attention = True if config_file['DecoderConfig']['attention'] == 'True' else False

        # generator
        self.beam_size = int(config_file['GeneratorConfig']['beam_size'])
        self.max_generate_len = int(config_file['GeneratorConfig']['max_generate_len'])

        # optimizer
        self.lr = float(config_file['OptimizerConfig']['lr'])
        self.weight_decay = float(config_file['OptimizerConfig']['weight_decay'])


class PreTrainedG2PModel(nn.Module):
    def __init__(self, config):
        super(PreTrainedG2PModel, self).__init__()
        self.config = config

    @classmethod
    def from_pretrained(cls, model_name):

        if model_name not in pretrained_models:
            raise ValueError

        # load config
        config = G2PConfig(pretrained_models[model_name].config_path)  # TODO add metod from_file

        # instantiate model
        model = cls(config)

        # loading weights
        state_dict = model_zoo.load_url(pretrained_models[model_name].url,
                                        progress=True, map_location=lambda storage, loc: storage)
        model.load_state_dict(state_dict)

        return model


class G2PModel(PreTrainedG2PModel):

    def __init__(self, config):
        super(G2PModel, self).__init__(config)

        # init
        self.config = config
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        # encoder
        self.encoder = Encoder(config)

        # decoder
        self.decoder = Decoder(config)
        self.attention = config.attention

        # generator
        self.beam_size = config.beam_size
        self.max_generate_len = config.max_generate_len

    def forward(self, x, x_length, y=None, p_length=None, n_best=1):
        # TODO rewrite desscription

        encoder_out, encoder_hc = self.encoder(x, x_length)

        if y is not None:
            out = self.decoder(y, p_length, encoder_hc, context=encoder_out if self.attention else None)
        else:
            out = self._generate(encoder_hc, context=encoder_out if self.attention else None)

        return out

    def _generate(self, hc, context=None):
        beam = Beam(self.config)
        h, c = hc
        # Make a beam_size batch.
        h = h.expand(h.size(1), beam.size, h.size(2))
        c = c.expand(c.size(1), beam.size, c.size(2))
        if context is not None:
            context = context.expand(beam.size, context.size(1), context.size(2))
        p_length = Variable(torch.from_numpy(np.array([1])))
        p_length = p_length.expand(beam.size).to(self.device)

        for i in range(self.max_generate_len):
            x = beam.get_current_state()
            o, hc = self.decoder(Variable(x.unsqueeze(1)).to(self.device), p_length, (h, c), context=context)
            if beam.advance(o.data.squeeze(1)):
                break
            h, c = hc
            h.data.copy_(h.data.index_select(1, beam.get_current_origin()))
            c.data.copy_(c.data.index_select(1, beam.get_current_origin()))
        return torch.LongTensor(beam.get_hyp(0)).unsqueeze(0)




import atexit
from datetime import datetime


_format = '%Y-%m-%d %H:%M:%S.%f'
_file = None
_run_name = None


def init(filename, run_name):
    global _file, _run_name
    _close_logfile()
    _file = open(filename, 'a')
    _file.write('\n-----------------------------------------------------------------\n')
    _file.write('Starting new training run\n')
    _file.write('-----------------------------------------------------------------\n')
    _run_name = run_name


def log(msg):
    print(msg)
    if _file is not None:
        _file.write(f'[{datetime.now().strftime(_format)[:-3]}] {msg}\n')


def _close_logfile():
  global _file
  if _file is not None:
    _file.close()
    _file = None


atexit.register(_close_logfile)



import torch.nn as nn
import torch.nn.functional as F
import torch

# TODO remove config, send par explicitly


class Encoder(nn.Module):

    def __init__(self, config):
        super(Encoder, self).__init__()

        self.vocab_size = config.encoder_vocab_size
        self.padding_idx = config.encoder_padding_idx

        self.d_embed = config.encoder_d_embed
        self.d_hidden = config.encoder_d_hidden
        self.num_layers = config.encoder_n_layers
        self.bidirectional = config.encoder_bidirectional

        self.embedding = nn.Embedding(self.vocab_size, self.d_embed, padding_idx=self.padding_idx)
        self.lstm = nn.LSTM(self.d_embed, self.d_hidden // 2 if self.bidirectional else self.d_hidden, self.num_layers,
                            batch_first=True, bidirectional=self.bidirectional)

    def forward(self, x, x_length):

        x = self.embedding(x)  # B x T x D
        x = nn.utils.rnn.pack_padded_sequence(x, x_length, batch_first=True, enforce_sorted=False)

        out, hc = self.lstm(x)

        out, _ = nn.utils.rnn.pad_packed_sequence(out, batch_first=True)

        if self.bidirectional:
            # stacking hidden and cell from forward and backward layers
            hc = tuple(torch.cat((hc_[0::2, :, :], hc_[1::2, :, :]), 2) for hc_ in hc)

        return out, hc


class Decoder(nn.Module):

    def __init__(self, config):
        super(Decoder, self).__init__()

        self.vocab_size = config.decoder_vocab_size
        self.padding_idx = config.decoder_padding_idx

        self.d_embed = config.decoder_d_embed
        self.d_hidden = config.decoder_d_hidden
        self.num_layers = config.decoder_n_layers

        self.embedding = nn.Embedding(self.vocab_size, self.d_embed, padding_idx=self.padding_idx)
        self.lstm = nn.LSTM(self.d_embed, self.d_hidden, self.num_layers, batch_first=True)
        if config.attention:
            self.attn = Attention(self.d_hidden)
        else:
            self.attn = None

        self.linear = nn.Linear(self.d_hidden, self.vocab_size)

    def forward(self, y, y_length, hc, context=None):

        batch_size, seq_len = y.size()

        y = self.embedding(y)  # B x T x D
        y = nn.utils.rnn.pack_padded_sequence(y, y_length, batch_first=True, enforce_sorted=False)

        out, hc = self.lstm(y, hc)

        out, _ = nn.utils.rnn.pad_packed_sequence(out, batch_first=True)

        if self.attn:
            out = self.attn(out, context)

        out = out.contiguous()

        out = self.linear(out.view(-1, out.size(2)))

        return F.log_softmax(out, dim=1).view(batch_size, -1, out.size(1)), hc


class Attention(nn.Module):
    """Dot global attention from https://arxiv.org/abs/1508.04025"""

    def __init__(self, dim):
        super(Attention, self).__init__()
        self.linear = nn.Linear(dim * 2, dim, bias=False)

    def forward(self, x, context):

        batch_size, seq_len, _ = x.size()

        attn = F.softmax(x.bmm(context.transpose(1, 2)), dim=2)
        weighted_context = attn.bmm(context)

        o = self.linear(torch.cat((x, weighted_context), 2).view(batch_size * seq_len, -1))
        return torch.tanh(o).view(batch_size, seq_len, -1)


class Beam(object):
    """Ordered beam of candidate outputs"""

    def __init__(self, config):
        """Initialize params"""
        self.size = config.beam_size
        self.done = False
        self.pad = config.decoder_padding_idx
        self.bos = config.decoder_bos_idx
        self.eos = config.decoder_eos_idx
        self.tt = torch.cuda if config.use_cuda else torch

        # The score for each translation on the beam.
        self.scores = self.tt.FloatTensor(self.size).zero_()

        # The backpointers at each time-step.
        self.prevKs = []

        # The outputs at each time-step.
        self.nextYs = [self.tt.LongTensor(self.size).fill_(self.pad)]
        self.nextYs[0][0] = self.bos

    # Get the outputs for the current timestep.
    def get_current_state(self):
        """Get state of beam."""
        return self.nextYs[-1]

    # Get the backpointers for the current timestep.
    def get_current_origin(self):
        """Get the backpointer to the beam at this step."""
        return self.prevKs[-1]

    def advance(self, workd_lk):
        """Advance the beam."""
        num_words = workd_lk.size(1)

        # Sum the previous scores.
        if len(self.prevKs) > 0:
            beam_lk = workd_lk + self.scores.unsqueeze(1).expand_as(workd_lk)
        else:
            beam_lk = workd_lk[0]

        flat_beam_lk = beam_lk.view(-1)

        best_scores, best_scores_id = flat_beam_lk.topk(self.size, 0, True, True)
        self.scores = best_scores

        # best_scores_id is flattened beam x word array, so calculate which
        # word and beam each score came from
        prev_k = best_scores_id // num_words
        self.prevKs.append(prev_k)
        self.nextYs.append(best_scores_id - prev_k * num_words)
        # End condition is when n-best are EOS.
        if self.nextYs[-1][0] == self.eos:
            self.done = True
        return self.done

    def get_hyp(self, k):
        """Get hypotheses."""
        hyp = []
        for j in range(len(self.prevKs) - 1, -1, -1):
            hyp.append(self.nextYs[j + 1][k])
            k = self.prevKs[j][k]
        return hyp[::-1]



import argparse

import torch
from torch.utils.data import DataLoader
from tqdm import tqdm

from ukro_g2p.datasets.lexicon_datasets import LexiconDataset, _collate_fn
from ukro_g2p.models.g2p_model import G2PConfig, G2PModel
from ukro_g2p.utils.util import phoneme_error_rate

from pathlib import Path
from ukro_g2p.utils.util import dump


def main(model, dataset, resume, out_dir):

    # init paths
    out_dir = Path(out_dir)
    ref_path = out_dir / 'ref'
    hyp_path = out_dir / 'hyp'
    scores_path = out_dir / 'scores'

    # make dir and files
    out_dir.mkdir(parents=True, exist_ok=True)
    try:
        ref_path.unlink()
        hyp_path.unlink()
        scores_path.unlink()
    except:
        pass

    # setup data_loader instances
    data_loader = DataLoader(dataset, batch_size=1, shuffle=False, collate_fn=_collate_fn, num_workers=1)

    # load model weights
    checkpoint = torch.load(resume, map_location=lambda storage, loc: storage)
    model.load_state_dict(checkpoint)

    # prepare model for testing
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    model = model.to(device)
    model.eval()

    # init mertic values
    avr_per = 0
    total_phonemes_length = 0
    num_batches = len(data_loader)

    with torch.no_grad():
        for batch_idx, batch in tqdm(enumerate(data_loader), total=num_batches):
            graphemes, graphemes_length, phonemes, _ = batch

            graphemes = graphemes.to(device)
            graphemes_length = graphemes_length.to(device)
            phonemes = phonemes.to(device)

            phonemes_predictions = model(graphemes, graphemes_length).tolist()
            phonemes_targets = phonemes[:, 1:].contiguous().tolist()

            for predictions, targets, input_graphemes in zip(phonemes_predictions, phonemes_targets, graphemes):
                targets_length = len(targets)
                per = phoneme_error_rate(predictions, targets)
                avr_per += per * targets_length
                total_phonemes_length += targets_length

                # saving hyp, ref and scores
                graphemes_str = ''.join([g for g in dataset.idx2graphemes(input_graphemes.tolist()) if g != '<s>'])
                predictions_str = ' '.join([p for p in dataset.idx2phonemes(predictions) if p not in ['<os>', '</os>']])
                targets_str = ' '.join([p for p in dataset.idx2phonemes(targets) if p not in ['<os>', '</os>']])
                dump(f'{graphemes_str}\t{predictions_str}', hyp_path, append=True)
                dump(f'{graphemes_str}\t{targets_str}', ref_path, append=True)
                dump(f'{round(per, 4)}\t{targets_str}\t{predictions_str}', scores_path, append=True)

    avr_per /= total_phonemes_length
    log_str = f'Phoneme Error Rate: {round(avr_per, 4)}'
    dump(log_str, scores_path, append=True)
    print(log_str)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--config', default=None, type=str,
                        help='config file path (default: None)')
    parser.add_argument('-d', '--dataset', default=None, type=str,
                        help='path to the dataset dir (default: None)')
    parser.add_argument('-r', '--resume', default=None, type=str,
                        help='path to latest checkpoint (default: None)')
    parser.add_argument('-o', '--out_dir', default=None, type=str,
                        help='path to the output dir (default: None)')

    args = parser.parse_args()

    config = G2PConfig(args.config)
    model = G2PModel(config)
    dataset = LexiconDataset(args.dataset, split='test')

    main(model, dataset, args.resume, args.out_dir)



import argparse

from ukro_g2p.datasets.lexicon_datasets import LexiconDataset
from ukro_g2p.trainer import Trainer
from torch.optim import Adam
import torch.nn as nn

from ukro_g2p.models.g2p_model import G2PConfig, G2PModel

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--config', default=None, type=str,
                        help='config file path (default: None)')
    parser.add_argument('-d', '--dataset', default=None, type=str,
                        help='path to the dataset dir (default: None)')
    parser.add_argument('-n', '--experiment_name', default=None, type=str,
                        help='path to the output dir (default: None)')
    parser.add_argument('-r', '--restore_epoch', default=-1, type=int,
                        help='restore epoch number (default: -1)')

    args = parser.parse_args()

    config = G2PConfig(args.config)
    model = G2PModel(config)
    optimizer = Adam(model.parameters(), lr=config.lr, weight_decay=config.weight_decay)
    loss = nn.NLLLoss(ignore_index=config.decoder_padding_idx)
    datasets = {'train': LexiconDataset(args.dataset, split='train'),
                  'dev': LexiconDataset(args.dataset, split='dev')}

    trainer = Trainer(model, datasets, optimizer, loss, epochs=100, batch_size=256,
                      experiment_name=args.experiment_name,
                      logging_freq=10, restore_epoch=args.restore_epoch)
    trainer.train_and_validate()



from pathlib import Path
from ukro_g2p.models.g2p_model import G2PConfig
from collections import namedtuple

tokenizer_obj = namedtuple("tokenizer_obj", ["config_path"])

pretrained_tokenizers = {
    "ukro-base-uncased": tokenizer_obj(
        config_path=Path(__file__).parent / "configs/ukro_base_uncased.config",
    )
}


class G2PTokenizer(object):
    def __init__(self, config):

        self.g2idx = {g: idx for idx, g in enumerate(config.graphemes)}
        self.idx2p = {idx: p for idx, p in enumerate(config.phonemes)}
        self.idx2h = {idx: h for idx, h in enumerate(config.human_phonemes)}

    def tokenize_graphemes(self, word):
        return list(word)

    def convert_graphemes_to_ids(self, graphemes):
        return [self.g2idx[g] for g in graphemes]

    def convert_ids_to_phonemes(self, ids):
        return [self.idx2p[i] for i in ids if self.idx2p[i] not in ['<os>', '</os>']]

    def convert_ids_to_human_phonemes(self, ids):
        return [self.idx2h[i] for i in ids if self.idx2h[i] not in ['<os>', '</os>']]

    @classmethod
    def from_pretrained(cls, tokenizer_name):

        if tokenizer_name not in pretrained_tokenizers:
            raise ValueError

        # load config
        config = G2PConfig(pretrained_tokenizers[tokenizer_name].config_path)  # TODO add metod from_file
        tokenizer = cls(config)

        return tokenizer



from ukro_g2p.utils.util import load_list


def read_lexicon_dataset(lexicon_dataset_path):
    """
    Read lexicon out of file to the dict in format - "id word pronunciation"
    :param lexicon_dataset_path: path to the lexicon dataset text file
    :return: lexicon dictionary
    """

    lexicon = {x.split()[0]: ' '.join(x.split()[1:]) for x in load_list(lexicon_dataset_path)}

    return lexicon



from torch.utils.data import Dataset
from pathlib import Path
from ukro_g2p.utils.lexicon_util import read_lexicon_dataset
from ukro_g2p.utils.util import load_list
import torch
import numpy as np

# TODO put this into config
GRAPHEMES_PADDING = 37
PHONEMES_PADDING = 99


class LexiconDataset(Dataset):
    def __init__(self, data_root, split='train'):
        self.data_root = Path(data_root)
        self.lexicon_dataset_path = self.data_root / 'data.txt'
        if split in ['train', 'dev', 'test']:
            self.data_keys_path = self.data_root / f'{split}_set'
        else:
            raise KeyError
        #TODO take it from config file
        self.phonemes_path = self.data_root / 'phones'
        self.letters_path = self.data_root / 'letters'

        self.lexicon = read_lexicon_dataset(self.lexicon_dataset_path)
        self.data_keys = load_list(self.data_keys_path)
        self.phoneme2idx = {p.split()[1]: int(p.split()[0]) for p in load_list(self.phonemes_path)}
        self.phoneme2idx['pad'] = len(self.phoneme2idx)
        self.phoneme2idx['<os>'] = len(self.phoneme2idx)
        self.phoneme2idx['</os>'] = len(self.phoneme2idx)
        self.idx2phoneme = dict((v, k) for k, v in self.phoneme2idx.items())
        self.grapheme2idx = {p.split()[1]: int(p.split()[0]) for p in load_list(self.letters_path)}
        self.grapheme2idx['pad'] = len(self.grapheme2idx)
        self.grapheme2idx['<s>'] = len(self.grapheme2idx)
        self.grapheme2idx['</s>'] = len(self.grapheme2idx)
        self.idx2grapheme = dict((v, k) for k, v in self.grapheme2idx.items())

    def __getitem__(self, idx):
        key = self.data_keys[idx]
        datapoint = self.lexicon[key]

        graphemes = datapoint.split()[0]
        phonemes = datapoint.split()[1:]

        graphemes_idx = self.graphemes2idx(graphemes)
        phonemes_idx = self.phonemes2idx(phonemes)
        return graphemes_idx, phonemes_idx

    def __len__(self):
        return len(self.data_keys)

    def graphemes2idx(self, graphemes):
        # graphemes_idx = [self.grapheme2idx['<s>']]
        graphemes_idx = [self.grapheme2idx[g] for g in graphemes]
        return graphemes_idx

    def phonemes2idx(self, phonemes):
        phonemes_idx = [self.phoneme2idx['<os>']]
        phonemes_idx.extend([self.phoneme2idx[p] for p in phonemes])
        phonemes_idx.append(self.phoneme2idx['</os>'])
        return phonemes_idx

    def idx2graphemes(self, idx):
        return [self.idx2grapheme[g] for g in idx]

    def idx2phonemes(self, idx):
        return [self.idx2phoneme[p] for p in idx]


def _collate_fn(batch):
    """
    Merges list of samples to form a mini-batch.
    Pads input sequences to longest inputs sequence in the batch
    Pads all target sequences to longest sequence in mini-batch with constraint

    Args:
        batch:

    Returns:

    """

    graphemes_lengths = list(map(lambda x: len(x[0]), batch))
    phonemes_lengths = list(map(lambda x: len(x[1]), batch))
    max_word_length = max(graphemes_lengths)
    max_phonemes_length = max(phonemes_lengths)
    batch_size = len(batch)

    graphemes_inputs = torch.LongTensor(batch_size, max_word_length).zero_() + GRAPHEMES_PADDING
    phonemes_targets = torch.LongTensor(batch_size, max_phonemes_length).zero_() + PHONEMES_PADDING
    graphemes_length = torch.LongTensor(batch_size).zero_()
    phonemes_length = torch.LongTensor(batch_size).zero_()

    for x in range(batch_size):
        sample = batch[x]
        graphemes, phonemes = sample

        graphemes = torch.from_numpy(np.asarray(graphemes)).long()

        phonemes = torch.from_numpy(np.asarray(phonemes)).long()

        graphemes_seq_length = graphemes.size(0)
        phonemes_seq_length = phonemes.size(0)

        graphemes_inputs[x].narrow(0, 0, graphemes_seq_length).copy_(graphemes)
        graphemes_length[x] = graphemes_seq_length
        phonemes_length[x] = phonemes_seq_length

        phonemes_targets[x].narrow(0, 0, phonemes_seq_length).copy_(phonemes)

    return graphemes_inputs, graphemes_length, phonemes_targets, phonemes_length



from setuptools import setup, find_packages, Command
import os
from os import path
from typing import Tuple, List
from shutil import rmtree
import sys


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options: List[Tuple] = []

    @staticmethod
    def status(s):
        """Print things in bold."""
        print(s)

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds...")
            rmtree(os.path.join(this_directory, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution...")
        os.system(f"{sys.executable} setup.py sdist bdist_wheel --universal")

        self.status("Uploading the package to PyPI via Twine...")
        os.system("twine upload dist/*")

        sys.exit()


setup(
    name="ukro_g2p",
    version="0.1.5",
    author="Kostiantyn Pylypenko",
    author_email="k.pylypenko@hotmail.com",
    description="NN based grapheme to phoneme model for Ukrainian language",
    license="MIT",
    keywords="Ukrainian grapheme to phoneme",
    url="https://github.com/kosti4ka/ukro_g2p",
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    cmdclass={"upload": UploadCommand},
)



from pathlib import Path
import json
from collections import OrderedDict
import Levenshtein


def read_json(fname):
    fname = Path(fname)
    with fname.open('rt') as handle:
        return json.load(handle, object_hook=OrderedDict)


def load_list(fname):
    fname = Path(fname)
    return [x for x in open(fname, 'r', encoding='utf-8').read().split('\n') if x]


def phoneme_error_rate(p_seq1, p_seq2):
    p_vocab = set(p_seq1 + p_seq2)
    p2c = dict(zip(p_vocab, range(len(p_vocab))))
    c_seq1 = [chr(p2c[p]) for p in p_seq1]
    c_seq2 = [chr(p2c[p]) for p in p_seq2]
    return Levenshtein.distance(''.join(c_seq1), ''.join(c_seq2)) / len(c_seq2)


def dump(iterable, file_name, append=False):

    # init paths and make sure the dir to write in exists
    file_name = Path(file_name)
    out_dir = file_name.parent
    out_dir.mkdir(parents=True, exist_ok=True)

    with open(file_name, 'a' if append else 'w', encoding='utf-8') as f:
        f.writelines(('\t'.join(str(ll) for ll in l) if type(l) != str and hasattr(l, '__iter__') else str(l)).rstrip('\r\n') + '\n' for l in ([iterable] if type(iterable) == str else iterable))
